// Implementation of chaskey 8 for cortex M0
//
// B. Haase
// public domain.
//
// gnu assembler format.
//

	.cpu cortex-m0
	.fpu softvfp
	.eabi_attribute 20, 1
	.eabi_attribute 21, 1
	.eabi_attribute 23, 3
	.eabi_attribute 24, 1
	.eabi_attribute 25, 1
	.eabi_attribute 26, 1
	.eabi_attribute 30, 2
	.eabi_attribute 34, 0
	.eabi_attribute 18, 4
		
	.text
	.align	2

	.align	2

	.global	chaskey8_aligned_asm
	.code	16
	.thumb_func
	.type	chaskey8_aligned_asm, %function

chaskey8_aligned_asm:
// #######################
// Chaskey 8 Round.:
// #######################
// chaskey 
// Implements the interface :
// uint64_t M0_chaskey (uint32_t *pIn, uint32_t len,uint32_t *pKey);
// 
// The first input byte shall be aligned to a 32 bit boundary,
// as imposed by the API, but odd or zero lengths are suppored.
// 
// pKey is expected to point to a 12-word array, where the first four
// words contain k, the next 4 words k1 = TWO(k) and the next four k2 = TWO(k1).
// 
// Actually the function returns full 128 bits of tags in r0,r1,r2,r3.
// This could be used with arm EABI by replacing the uint64_t with an 
// 'aggregate' type of four 32 bit words. 
//
// Author: B. Haase, Public domain. 
    push {r4,r5,r6,r7,r14}
    .syntax unified
    mov r4,r8
    .syntax divided
    .syntax unified
    mov r5,r9
    .syntax divided
    push {r2,r4,r5}
    ldm r2!,{r4,r5,r6,r7}
    sub r1,#16
    bgt nextMessageBlock
    b lastMessageBlock
nextMessageBlock:
    .syntax unified
    mov r8,r0
    .syntax divided
    .syntax unified
    mov r9,r1
    .syntax divided
    ldm r0,{r0,r1,r2,r3}
    eor r4,r0
    eor r5,r1
    eor r6,r2
    eor r7,r3
// chaskey permute 
// expects input of state in r4 (v0),r5 (v1),r6 (v2),r7 (v3)
// clobbers r0 ... r3 
    mov r0,#16
    mov r1,#27
    mov r2,#24
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
// end chaskey permute 
    .syntax unified
    mov r0,r8
    .syntax divided
    add r0,#16
    .syntax unified
    mov r1,r9
    .syntax divided
    sub r1,#16
    ble lastMessageBlock
    b nextMessageBlock
lastMessageBlock:
    beq fullLastMessageBlock
    add r1,#16
    mov r2,#0
    mov r3,#0
    push {r2,r3}
    push {r2,r3}
    .syntax unified
    mov r2,r13
    .syntax divided
nextByte:
    sub r1,#1
    blt paddingByte
    ldrb r3,[r0,#0]
    add r0,#1
    strb r3,[r2,#0]
    add r2,#1
    b nextByte
paddingByte:
    mov r3,#1
    strb r3,[r2,#0]
    pop {r0,r1,r2,r3}
    eor r4,r0
    eor r5,r1
    eor r6,r2
    eor r7,r3
    mov r3,#32
    b lastPermutation
fullLastMessageBlock:
    ldm r0,{r0,r1,r2,r3}
    eor r4,r0
    eor r5,r1
    eor r6,r2
    eor r7,r3
    mov r3,#16
lastPermutation:
    pop {r2}
    add r2,r3
    .syntax unified
    mov r8,r2
    .syntax divided
    ldm r2,{r0,r1,r2,r3}
    eor r4,r0
    eor r5,r1
    eor r6,r2
    eor r7,r3
// chaskey permute 
// expects input of state in r4 (v0),r5 (v1),r6 (v2),r7 (v3)
// clobbers r0 ... r3 
    mov r0,#16
    mov r1,#27
    mov r2,#24
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
    add r4,r5
    ror r5,r1
    eor r5,r4
    ror r4,r0
    add r6,r7
    ror r7,r2
    eor r7,r6
    add r4,r7
    mov r3,#19
    ror r7,r3
    eor r7,r4
    add r6,r5
    mov r3,#25
    ror r5,r3
    eor r5,r6
    ror r6,r0
// end chaskey permute 
    .syntax unified
    mov r0,r8
    .syntax divided
    ldm r0,{r0,r1,r2,r3}
    eor r0,r4
    eor r1,r5
    eor r2,r6
    eor r3,r7
    pop {r4,r5}
    .syntax unified
    mov r8,r4
    .syntax divided
    .syntax unified
    mov r9,r5
    .syntax divided
    pop {r4,r5,r6,r7,r15}
//Instruction count of chaskey (Cortex M0): 330 instructions.

